//
//  ES1Renderer.m
//  ARsample
//
//  Created by hak on 5/14/10.
//  Copyright __MyCompanyName__ 2010. All rights reserved.
//

#import "ES1Renderer.h"

@implementation ES1Renderer

// Create an OpenGL ES 1.1 context
- (id)init
{
    if ((self = [super init]))
    {
        context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];

        if (!context || ![EAGLContext setCurrentContext:context])
        {
            [self release];
            return nil;
        }

        // Create default framebuffer object. The backing will be allocated for the current layer in -resizeFromLayer
        glGenFramebuffersOES(1, &defaultFramebuffer);
        glGenRenderbuffersOES(1, &colorRenderbuffer);
        glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
        glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
        glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, colorRenderbuffer);
    }

	perspective( /* field of view in degree */ 40.0,
				/* aspect ratio */ 0.666,
				/* Z near */ VIEWPORT_NEAR,
				/* Z far */ VIEWPORT_FAR);
    return self;
}

- (void)render
{
    static const GLfloat squareVertices2[] = {
        -0.5f,  -0.5f, 10.f,
        0.5f,  -0.5f,  10.f,
        -0.5f,   0.5f, 10.f,
        0.5f,   0.5f,  10.f,
    };

    static const GLubyte squareColors[] = {
        255, 255,   0, 255,
        0,   255, 255, 255,
        0,     0,   0,   0,
        255,   0, 255, 255,
    };

    // This application only creates a single context which is already set current at this point.
    // This call is redundant, but needed if dealing with multiple contexts.
    [EAGLContext setCurrentContext:context];

    // This application only creates a single default framebuffer which is already bound at this point.
    // This call is redundant, but needed if dealing with multiple framebuffers.
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
    glViewport(0, 0, backingWidth, backingHeight);

    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT);

    //Set perspective
    perspective( /* field of view in degree */ 40.0,
				/* aspect ratio */ 0.666,
				/* Z near */ VIEWPORT_NEAR,
				/* Z far */ VIEWPORT_FAR);
    
    float fAzimuth = [((ARsampleAppDelegate*)[UIApplication sharedApplication].delegate) _azimuth];
    float fAccelerometerX = [((ARsampleAppDelegate*)[UIApplication sharedApplication].delegate) _accerelometerX];
    float fAccelerometerY = [((ARsampleAppDelegate*)[UIApplication sharedApplication].delegate) _accerelometerY];
    float fAccelerometerZ = [((ARsampleAppDelegate*)[UIApplication sharedApplication].delegate) _accerelometerZ];
    float fMagnetometerX = [((ARsampleAppDelegate*)[UIApplication sharedApplication].delegate) _magnetometerX];
    float fMagnetometerY = [((ARsampleAppDelegate*)[UIApplication sharedApplication].delegate) _magnetometerY];
    float fMagnetometerZ = [((ARsampleAppDelegate*)[UIApplication sharedApplication].delegate) _magnetometerZ];

    //
    //Retrieve Tilt value
    float fTilt;
	if (fAccelerometerY > 0.f) {
		fTilt = atanf( fAccelerometerX / fAccelerometerY );
	}
	else if (fAccelerometerY < 0.f) {
		fTilt = atanf( fAccelerometerX / fAccelerometerY ) + M_PI;
	}
    else if (fAccelerometerX < 0) {
		fTilt =  0;
	}
    else
    {
		fTilt =  M_PI;
	} 
    fTilt = fmodf(fTilt + 2 * M_PI, 2 * M_PI);    
    fAzimuth = fmod( fAzimuth + fTilt + 2 * M_PI, 2 * M_PI);

    //
    //Retrieve pitch value
    float f;
    float fPitch;
    if( fabs( fAccelerometerY ) > fabs( fAccelerometerX ) )
    {
        f = fAccelerometerY;        
    }
    else
    {
        f = -fAccelerometerX;        
    }
    
    if (fAccelerometerZ > 0.f) {
		fPitch = atanf( f / fAccelerometerZ );
	}
	else if (fAccelerometerZ < 0.f) {
		fPitch = atanf( f / fAccelerometerZ ) + M_PI ;
	}
    else if (f < 0) {
		fPitch = 0;
	}
    else
    {
		fPitch = M_PI;
	} 
    fPitch = fmodf(fPitch + 2 * M_PI + M_PI / 2, 2 * M_PI);    

#ifdef DEBUG
    NSLog([NSString stringWithFormat:@"Y%f,P%f,R%f MX:%f,MY:%f,MZ:%f AX:%f,AY:%f,AZ:%f \n",
           -fAzimuth,
           fPitch,
           fTilt,
           fMagnetometerX,
           fMagnetometerY,
           fMagnetometerZ,
           fAccelerometerX,
           fAccelerometerY,
           fAccelerometerZ
           ]);
#endif
    
    //Set Camera
    setCameraYPR( -fAzimuth, fPitch, fTilt );

    //Render scene
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, squareColors);
    glEnableClientState(GL_COLOR_ARRAY);    
    glVertexPointer(3, GL_FLOAT, 0, squareVertices2);
    glEnableClientState(GL_VERTEX_ARRAY);

    float fAngle = 360 / 16;
    for( int32_t i = 0; i < 16; ++ i )
    {
        glPushMatrix();
        glRotatef( fAngle * i, 0.0f, 1.0f, 0.0f);
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);    
        glPopMatrix();
    }

    // This application only creates a single color renderbuffer which is already bound at this point.
    // This call is redundant, but needed if dealing with multiple renderbuffers.
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context presentRenderbuffer:GL_RENDERBUFFER_OES];
}

- (BOOL)resizeFromLayer:(CAEAGLLayer *)layer
{	
    // Allocate color buffer backing based on the current layer size
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:layer];
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &backingWidth);
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight);

    if (glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES)
    {
        NSLog(@"Failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
        return NO;
    }

    return YES;
}

- (void)dealloc
{
    // Tear down GL
    if (defaultFramebuffer)
    {
        glDeleteFramebuffersOES(1, &defaultFramebuffer);
        defaultFramebuffer = 0;
    }

    if (colorRenderbuffer)
    {
        glDeleteRenderbuffersOES(1, &colorRenderbuffer);
        colorRenderbuffer = 0;
    }

    // Tear down context
    if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];

    [context release];
    context = nil;

    [super dealloc];
}

@end
